"""JSON Web Token (JWT) encoding and decoding library for Python 2 and 3.
Extends pyjwt to include Atlassian's custom query string hash (qsh) claim.
"""

from .encode import encode_token
from .authenticate import Authenticator, DecodeError



"""Authenticate an HTTP request that contains an Atlassian-style JWT token.

Atlassian-style JWT tokens include a `qsh` claim which stands for query string
hash. See `Understanding JWT for Atlassian Connect`_ for more details.

.. Understanding JWT for Atlassian Connect
   _https://developer.atlassian.com/blog/2015/01/understanding-jwt/
"""

import abc
import jwt as jwt
from jwt import DecodeError
from .url_utils import hash_url, parse_query_params

class Authenticator(object):
    """An abstract base class for authenticating Atlassian Connect requests.

    Subclasses *must* implement the `get_shared_secret` method.

    Example:
        Subclass this abstract base class to provide authentication to an
        Atlassian Connect Addon.

        import atlassian_jwt

        class MyAddon(atlassian_jwt.Authenticator):
            def __init__(self, tenant_info_store):
                super(MyAddon, self).__init__()
                self.tenant_info_store = tenant_info_store

            def get_shared_secret(self, client_key):
                tenant_info = self.tenant_info_store.get(client_key)
                return tenant_info['sharedSecret']

        my_auth = MyAddon(tenant_info_store)
        try:
            client_key = my_auth.authenticate(http_method, url, headers)
            # authentication succeeded
        except atlassian_jwt.DecodeError:
            # authentication failed
            pass
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, algorithms=('HS256',), leeway=10):
        self.algorithms = algorithms
        self.leeway = leeway

    @abc.abstractmethod
    def get_shared_secret(self, client_key):
        """Get the shared secret associtated with the client key.

        Subclasses of this abstract base class *must* implement this method.

        Use the client key to retrieve the shared secret (presumably) from a
        persistent store of which this abstract base class does not need to
        know the details.

        This is the shared secret that was used to sign the JWT token and can
        be used to verify its authenticity.

        Args:
            client_key (string): client key

        Returns:
            string: shared secret used to sign the JWT token
        """
        raise NotImplementedError

    def authenticate(self, http_method, url, headers=None):
        """Extract the JWT token from the `Authorization` header, or if not
        found there then the `jwt` query parameter.

        Args:
            http_method (string): HTTP method

            url (string): URL

            headers (dict): incoming request headers. The header name
                `Authorization` is case-insensitive. The token type `JWT` in
                the `Authorization` header is case-insensitive.

        Returns:
            string: client key (the `iss` claim from the JWT token)

        Raises:
            DecodeError: If neither `headers` nor the query parameteters in
                `url` contain a JWT token. Or if `qsh` claim does not match
                expected value.

        .. _Exposing a service:
           https://developer.atlassian.com/static/connect/docs/latest/concepts/authentication.html#exposing    
        """
        token = self._get_token(
            headers=headers,
            query_params=parse_query_params(url))

        claims = jwt.decode(token, verify=False, algorithms=self.algorithms,
                            options={"verify_signature": False})
        if claims['qsh'] != hash_url(http_method, url):
            raise DecodeError('qsh does not match')

        # verify shared secret
        jwt.decode(
            token,
            audience=claims.get('aud'),
            key=self.get_shared_secret(claims['iss']),
            algorithms=self.algorithms,
            leeway=self.leeway)

        # return client key
        return claims['iss']

    @staticmethod
    def _get_token(headers=None, query_params=None):
        if headers:
            for name, value in headers.items():
                if name.lower() == 'authorization':
                    parts = value.split()
                    if len(parts) > 1 and parts[0].lower() == 'jwt':
                        return parts[1]

        if query_params and 'jwt' in query_params:
            value = query_params['jwt']
            return value if isinstance(value, str) else value[0]

        raise DecodeError('JWT token not found')



"""Helper functions for dealing with URLs.
"""
from hashlib import sha256

try:
    from urllib.parse import quote, urlparse, parse_qs
except ImportError:
    from urllib import quote
    from urlparse import urlparse, parse_qs


def hash_url(http_method, url):
    """Hash a URL to create a value suitable for a `qsh` claim.

    See `Creating a query string hash`_ for more info.

    Args:
        http_method (string): HTTP method that will be used for the ensuing
            request.
        url (string): URL to hash. Must be relative to the host base URL found
            in the tenant information.

    Returns:
        string: suitable to be used as the value of a `qsh` claim in a JWT
            token.

    .. _Creating a query string hash:
       https://developer.atlassian.com/static/connect/docs/latest/concepts/understanding-jwt.html#qsh    
    """
    canonical_request = canonicalize_request(http_method, url)
    return sha256(canonical_request.encode('utf8')).hexdigest()


def canonicalize_request(http_method, url):
    """Form a `canonical-request` as defined in `Creating a query string hash`_

    Example:
        Give then following request

            GET "/path/to/service?zee_last=param&repeated=parameter 1&first=param&repeated=parameter 2"

        the canonical request is

            GET&/path/to/service&first=param&repeated=parameter%201,parameter%202&zee_last=param

    Args:
        http_method (string): HTTP method that will be used for the ensuing
            request.
        url (string): URL to hash. Must be relative to the host base URL found
            in the tenant information.

    Returns:
        string: canonical request

    .. _Creating a query string hash:
       https://developer.atlassian.com/static/connect/docs/latest/concepts/understanding-jwt.html#qsh
    """

    # This code ensures that in python 2 if you pass a unicode string with ascii
    # characters that it will be converted into a str. If you pass a unicode
    # str with non-ascii characters then its not a valid URL so we just error
    # out.
    #
    # The python URL libraries only support parsing and encoding URLs when
    # passed in as a str. In Python 2 this is an ascii string and in Python 3
    # its unicode.
    #
    # This does't mean our code doesn't support unicode. URLs must have i18n
    # characters encoded as ASCII using either Punycode or percent encoding.
    # We only accept valid encoded URLs, that is, URLs you could actually send
    # in a HTTP request.
    try:
        # Python 2 has str and unicode strings.
        if isinstance(url, unicode):
            url = url.encode()
    except NameError:
        # Python 3.
        pass

    # We want to treat '//path' as a path, not a hostname, so we must strip
    # out leading slashes before using urlparse:
    parsed_url = urlparse(url.lstrip('/'))

    def canonicalize_uri():
        path = parsed_url.path.strip('/').replace('&', quote('&'))
        yield '/' + path
        if parsed_url.params:
            # path params (not to be confused with query params)
            yield ';' + parsed_url.params

    def canonicalize_qs():
        parameters = parse_qs(parsed_url.query, keep_blank_values=True)
        for k, v in parameters.items():
            if k == 'jwt':
                continue
            # This incorrectly sorts the values by their `unquoted` values. It
            # should be sorting by the encoded values but the AC implementation
            # has a bug (JWT-20).
            yield quote(k, '~'), ','.join([quote(s, '~') for s in sorted(v)])

    canonical_uri = ''.join(canonicalize_uri())
    canonical_qs = '&'.join('='.join(x) for x in sorted(canonicalize_qs()))
    return '%s&%s&%s' % (http_method.upper(), canonical_uri, canonical_qs)


def parse_query_params(url):
    """Parse query parameters from a URL's query string.

    Args:
        url (string): URL from which to parse query parameters

    Returns:
        dict: the query parameters. Dictionary keys are query parameter names.
            Dictionary values are lists of string query parameter values as the
            same query parameter name can be listed multiple time in a URL's
            query string.

    .. _Creating a query string hash:
       https://developer.atlassian.com/static/connect/docs/latest/concepts/understanding-jwt.html#qsh
    """    
    return parse_qs(urlparse(url).query)



from setuptools import setup

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = ''

setup(
    name = 'atlassian-jwt',
    packages = ['atlassian_jwt'],
    version = '1.9.0',
    description = 'JSON web token: pyjwt plus Atlassian query-string-hash claim',
    long_description=long_description,
    author = 'Atlassian',
    author_email = 'bedwards@atlassian.com',
    url = 'https://bitbucket.org/atlassian/atlassian-jwt-py',
    install_requires = [
        'PyJWT>=1.5.0'
    ],
    keywords = [
        'jwt',
        'json',
        'web',
        'token',
        'pyjwt',
        'atlassian',
        'connect',
        'addon',
        'query',
        'string',
        'hash',
        'qsh',
    ],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



"""Encode JWT tokens with Atlassian Connect `qsh` claim.

This module provides an `encode_token` function that can be used to encode a
JWT including the query string hash claim.
"""

from time import time
import jwt as jwt
from .url_utils import hash_url


def encode_token(http_method, url, clientKey, sharedSecret, timeout_secs=60*60,
                 **ignored):
    """Encode a JWT token including the `qsh` claim.

    Args:
        http_method (string): HTTP method used for the ensuing request.
        url (string): URL to sign. Must be relative to the host base URL found
            in the tenant information.
        clientKey (string): client key from tenant information used as `aud`
            claim in token. camel case to allow JSON tenant information (from
            installed lifecycle callback) to be passed in directly.
        sharedSecret (string): key used to encrypt the token. camel case to
            allow tenant information to be passed in directly.
        timeout_secs (Optional[int]): timeout for the generated token in
            seconds. Defaults to one hour.
        ignored (dict): allows tenant information to be passed in directly.

    Returns:
        string: the encoded JWT token
    """
    now = int(time())
    token = jwt.encode(key=sharedSecret, algorithm='HS256', payload={
        'aud': clientKey,
        'exp': now + timeout_secs,
        'iat': now,
        'iss': clientKey,
        'qsh': hash_url(http_method, url),
    })
    if isinstance(token, bytes):
        token = token.decode('utf8')
    return token
